--[[
Features:
LibQuickHealth-2.0 is a library that provides more up to date health data than the default Blizz events and functions.
In addition to the standard UNIT_HEALTH event, the library listens to combat log events (CLEU) to update health values. CLEU events fire more often than UNIT_HEALTH, and allow the library to get around the ~300ms throttling of UNIT_HEALTH updates.
The interesting CLEU events are those containing health differences (damage and heals). Normally, when UNIT_HEALTH fires, those differences add up to the value reported by Blizz's UnitHealth. But sometimes, when a unit takes damage according to the CLEU, that damage is processed *after* the next UNIT_HEALTH event. This is called "pending damage". One of the features of LibQuickHealth-2.0 (and an improvement over 1.0) is to detect pending damage, and incorporate it into the health values that are presented to players.

Limitations:
1.) Sometimes there are combat log entries like this: -6000 (damage), +5000 (heal). But what really happened on the server was that the heal occured first. If at least part of this heal was overheal, the prediction (-6000+5000 = -1000) is way off, because what happened is for example +50 heal first(4950 overheal), and then 6000 damage, for a total of +50-6000 = -5950. No AddOn could possibly prevent or predict this, but luckily these cases are very rare. 
However, if you see something like this happening (tank at ~100%, then takes damage instantly followed by a heal), you should know that this heal might have been overheal. So don't cancel your heal. ;)

2.) If the maximum health changes, pending damage is completely ignored. Example: The tank is at 11001/18000, takes 9001 damage and uses Last Stand at the same time. Your client might tell you now that the tank health is 16401/23400 (+30% health and max health), because that 9k hit was not yet processed (pending damage). What the library does is: Show the tank at 11001/18000, then at 11001-9001 = 2000/18000, then at 16401/23400, because pending damage is ignored. But on the server, the tank has only 16401-9001 = 7400/23400 health!
This looks very similar to the case before: The tank takes damage, but shortly afterwards that damage is (magically) undone. Again, if you see this, don't cancel your heal!

3.) Bloodrush. There's no combat log entry for the health lost. I believe the health cost depends on the caster level and race, but I don't know the exact formula.

4.) Exceptionally large HP5 values. There are no combat log entries for HP5 ticks, which occur every 2 seconds. The library assumes that the health regen in combat is between 0 and 50 HP5, i.e. HP5 ticks between 0 and 20. 
Common HP5 values: Demon skin/armor <=18 (24 talented?), Elixir of Major Fortitude 10, Enchant Boots - Vitality 4. Less common: Dreaming Glory 30 (buff from herbalism), and a few level 60 items. 
So the library might get confused by two things: Small amounts of pending damage (20 or less), and Warlocks with herbalism...

Usage:
To start listening to the event, do QuickHealth.RegisterCallback(YourAddonTableHere, "HealthUpdated", HandlerMethodHere);
The callback triggers with 5 arguments: self, event, guid, health, healthMax (first two passed by the callbackhandler)
Additionaly, QuickHealth:UnitHealth(unitID) can be called to get what QuickHealth thinks is the current health.

Example:
local QuickHealth = LibStub("LibQuickHealth-2.0")
local MyAddon = {}
function MyAddon:Initialize()
	-- Note: It's QuickHealth.RegisterCallback(), not QuickHealth:RegisterCallback().
	QuickHealth.RegisterCallback(MyAddon, "HealthUpdated", "HealthUpdated")
	-- Or:
	-- QuickHealth.RegisterCallback(MyAddon, "HealthUpdated", MyAddon.HealthUpdated)
end
function MyAddon:HealthUpdated(event, guid, health, healthMax)
	if event == "HealthUpdated" then
		ChatFrame1:AddMessage(format("%s HealthUpdated %i/%i", guid, health, healthMax))
		if guid==UnitGUID("player") then
			assert(health==LibQuickHealth:UnitHealth("player")
		end
	end
end

--]]

local MAJOR, MINOR = "LibQuickHealth-2.0", 1
local QuickHealth, oldminor = LibStub:NewLibrary(MAJOR, MINOR)

QuickHealth.events = QuickHealth.events or LibStub("CallbackHandler-1.0"):New(QuickHealth)

local healthFromGUID = {}
local healthMaxFromGUID = {}
local healthQueueFromGUID = {}

QuickHealth.eventHandlers = QuickHealth.eventHandlers or {}
local eventHandlers = QuickHealth.eventHandlers
local eventFrame
local initialized

-- Public API functions.
function QuickHealth:UnitHealth(unitID)
	local guid = UnitGUID(unitID)
	return guid and healthFromGUID[guid] or UnitHealth(unitID)
end

function QuickHealth:UnitHealthQueue(unitID)
	local guid = unitID
	if guid and not healthQueueFromGUID[guid] then
		guid = UnitGUID(unitID)
	end
	if guid then
		return unpack(healthQueueFromGUID[guid])
	end
end

-- Implementation.
if(QuickHealth.eventFrame) then
	eventFrame = QuickHealth.eventFrame
else
	eventFrame = CreateFrame("Frame")
	QuickHealth.eventFrame = eventFrame
	eventFrame:SetScript("OnEvent", function(self, event, ...) eventHandlers[event](...) end)
end

local HealthUpdatedRegistered
local HealthUpdatedDebugRegistered

local GetFrameNumber
do
	local counter = 0
	function GetFrameNumber() 
		return counter 
	end
	local function IncrementCounter()
		counter = counter+1 
	end

	-- When an addon starts to listen for HealthUpdated we register our events and actualy start doing something
	function QuickHealth.events:OnUsed(target, eventName)
		if eventName == "HealthUpdatedDebug" and not HealthUpdatedDebugRegistered then
			counter = 0
			eventFrame:SetScript("OnUpdate", IncrementCounter)
		end

		local register = eventName == "HealthUpdated" or eventName == "HealthUpdatedDebug"
		if register and not (HealthUpdatedRegistered or HealthUpdatedDebugRegistered) then
			eventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
			eventFrame:RegisterEvent("UNIT_MAXHEALTH")
			eventFrame:RegisterEvent("UNIT_HEALTH")
			eventFrame:RegisterEvent("PLAYER_UNGHOST")
		end
		if eventName == "HealthUpdated" then
			HealthUpdatedRegistered = true
		end
		if eventName == "HealthUpdatedDebug" then
			HealthUpdatedDebugRegistered = true
		end

		if not initialized then
			if UnitHealth('player') and eventHandlers.PLAYER_LOGIN then
				eventHandlers.PLAYER_LOGIN()
			else
				eventFrame:RegisterEvent("PLAYER_LOGIN")
			end
		end
	end

	-- Unregister the events and empty the healthtable
	function QuickHealth.events:OnUnused(target, eventName)
		if eventName == "HealthUpdated" then
			HealthUpdatedRegistered = nil
		end
		if eventName == "HealthUpdatedDebug" then
			HealthUpdatedDebugRegistered = nil
			eventFrame:SetScript("OnUpdate", nil)
		end
		if not (HealthUpdatedRegistered or HealthUpdatedDebugRegistered) then
			eventFrame:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
			eventFrame:UnregisterEvent("UNIT_MAXHEALTH")
			eventFrame:UnregisterEvent("UNIT_HEALTH")
			eventFrame:UnregisterEvent("PLAYER_UNGHOST")
		end
	end
end

local function clear(queue)
	for i=#queue,1,-1 do
		queue[i]=nil
	end
end

local MAX_QUEUE_SIZE = 5
local MAX_HP5_TICK = 20 -- 50 HP5
local function processUH(guid, unitHealth, unitHealthMax)
	if unitHealthMax == 100 then
		-- We might get here for unitID == target
		healthFromGUID[guid] = nil
		healthMaxFromGUID[guid] = nil
		healthQueueFromGUID[guid] = nil
		return
	end

	local health = healthFromGUID[guid]
	if not health then
		healthFromGUID[guid] = unitHealth
		healthMaxFromGUID[guid] = unitHealthMax
		healthQueueFromGUID[guid] = {0,0,0,0,0}
		clear(healthQueueFromGUID[guid])
		if HealthUpdatedRegistered then
			QuickHealth.events:Fire("HealthUpdated", guid, unitHealth, unitHealthMax)
		end
		return
	end

	local healthMax = healthMaxFromGUID[guid]
	local healthQueue = healthQueueFromGUID[guid]
	local fire
	if healthMax ~= unitHealthMax then
		-- TODO If healthMax changes, one of three things can happen:
		-- 1. health doesn't change (except if health>unitHealthMax ofc)
		-- 2. health changes by unitHealthMax-healthMax
		-- 3. health/healthMax doesn't change: h1/m1 = h2/m2, h2 = h1*m2/m1, h2-h1 = h1*(m2/m1-1)
		healthMax = unitHealthMax
		healthMaxFromGUID[guid] = unitHealthMax
		-- For now, just reset and hope it's correct. 
		health = unitHealth
		healthFromGUID[guid] = unitHealth
		fire = true
	end

	-- Ex unitHealth = 104, health = 100
	local discrepancy = unitHealth - health
	if discrepancy == 0 then
		clear(healthQueue)
	elseif #healthQueue==0 then
		health = unitHealth
		healthFromGUID[guid] = unitHealth
		fire = true
	elseif 0<discrepancy and discrepancy<=MAX_HP5_TICK then
		clear(healthQueue)
		health = unitHealth
		healthFromGUID[guid] = health
		fire = true
	else
		local noexplanation = true

		local numdamage = 0
		local pendingDamage
		local healthChanged
		do
			-- Case 1: The last damage entries are pending.
			-- Ex. health = 1000-200-100+250 = 950, healthMax = 1000, unitHealth = 1000, 
			-- queue = (-200, -100, +250) = (-200, +200, -100) -> (-100)
			local h=health
			local pending = 0
			local overheal = 0
			for i=#healthQueue, 1, -1 do
				local amount = healthQueue[i]
				if amount<0 then
					numdamage = numdamage+1
					-- We assume that amount is pending, so it didn't count yet.
					-- Ex. i=2: h = 1000-(-100) = 1100
					h = h - amount
					pending = pending + amount
					-- Check for possible OHs.
					if h > healthMax then
						-- Ex. i=2: h=1000
						overheal = overheal + h-healthMax
						h = healthMax
					end
					local diff = unitHealth - h
					if 0<=diff and diff<=MAX_HP5_TICK then
						noexplanation = nil
						if diff>0 or overheal>0 then
							-- If we get here, our explanation is that an HP5 tick occured.
							healthChanged = true
						end
						pendingDamage = pending
						break
					end
				end
			end
		end

		if noexplanation and numdamage>1 then
			-- Case 2: A single damage entry is pending, but not the last one.
			for i=#healthQueue, 1, -1 do
				local amount = healthQueue[i]
				if amount<0 then
					-- See what happens if amount is pending.
					local h = health - amount
					-- Check for OH.
					if h>healthMax then
						h = healthMax
					end
					local diff = unitHealth - h
					if 0<=diff and diff<=MAX_HP5_TICK then
						noexplanation = nil
						if diff>0 then
							-- If we get here, our explanation is that an HP5 tick occured.
							healthChanged = true
						end
						pendingDamage = amount
						break
					end
				end
			end
		end

		if pendingDamage then
			-- Remove everything except the pending damage.
			healthQueue[1] = pendingDamage
			for j=#healthQueue, 2, -1 do
				healthQueue[j]=nil
			end
			if healthChanged then
				health = unitHealth + pendingDamage
				healthFromGUID[guid] = health
				fire = true
			end
		end

		if noexplanation then
			-- Case 3: Nothing's pending, but the order of the combat log entries might have been wrong.
			-- If this results in more overheal than predicted, our prediction was wrong.
			-- We can't do anything about that. Just reset.
			-- Examples: 
			-- 19,179: (-a,+b,-c,+d) = (+b OH, -a, -c, +d) -> ()
			-- 19,1169: (-a,+b) = (+b OH,-a) -> ()
			-- 19,1426: (-a,-b,+c) = (-a,+c OH,-b) -> ()
			-- 19,1253: (-a pending, +b) = (+b OH, -a) -> ()
			-- 19,1942: (-a pending, +b,+c) = (+b,+c OH, -a) -> ()   [can't be (+b OH,-a,+c)]
			clear(healthQueue)
			health = unitHealth
			healthFromGUID[guid] = unitHealth
			fire = true
		end
	end

	if --[[fire and]] HealthUpdatedRegistered then
		QuickHealth.events:Fire("HealthUpdated", guid, health, healthMax)
	end
end

-- Amount is just the raw value from the combat log, including overheal.
local function processCLEU(guid, amount, amountMax)
	local health = healthFromGUID[guid]
	if not health then return end
	local healthQueue = healthQueueFromGUID[guid]
	local healthMax = healthMaxFromGUID[guid]

	if amountMax then
		healthMax = healthMax+amountMax
		healthMaxFromGUID[guid] = healthMax
	end

	local healthDiff = amount
	if health+amount>healthMax then
		healthDiff = healthMax-health
	end
	--print(healthDiff)

	if healthDiff~=0 then
		tinsert(healthQueue, healthDiff)
		if #healthQueue>MAX_QUEUE_SIZE then
			local first = table.remove(healthQueue, 1)
			healthQueue[1] = healthQueue[1]+first
		end
		health = health+healthDiff
		healthFromGUID[guid] = health
		if HealthUpdatedRegistered then
			QuickHealth.events:Fire("HealthUpdated", guid, health, healthMax)
		end
	end
end

eventHandlers.CLEU = processCLEU
eventHandlers.UH = processUH


function eventHandlers.PLAYER_LOGIN()
	for _,unitID in pairs({"player","pet"}) do
		if UnitHealth(unitID) and UnitGUID(unitID) then
			initialized = true
			eventHandlers.UNIT_HEALTH(unitID)
		end
	end
	if initialized then
		eventHandlers.PLAYER_LOGIN = nil
	end
end

function eventHandlers.PLAYER_UNGHOST()
	eventHandlers.UNIT_HEALTH("player")
end

function eventHandlers.UNIT_HEALTH(unitID)
	if unitID == 'target' or unitID == 'focus' or unitID == 'mouseover' then return end
	local unitHealth = UnitHealth(unitID)
	local unitHealthMax = UnitHealthMax(unitID)

	if HealthUpdatedDebugRegistered then
		QuickHealth.events:Fire("HealthUpdatedDebug", UnitGUID(unitID), 1, GetTime(), GetFrameNumber(), unitHealth, unitHealthMax)
	end

	processUH(UnitGUID(unitID), unitHealth, unitHealthMax)
end

function eventHandlers.UNIT_MAXHEALTH(unitID)
	if unitID == 'target' or unitID == 'focus' or unitID == 'mouseover' then return end
	local unitHealth = UnitHealth(unitID)
	local unitHealthMax = UnitHealthMax(unitID)
	if HealthUpdatedDebugRegistered then
		QuickHealth.events:Fire("HealthUpdatedDebug", UnitGUID(unitID), 2, GetTime(), GetFrameNumber(), unitHealth, unitHealthMax)
	end

	processUH(UnitGUID(unitID), unitHealth, unitHealthMax)
end

local lifeTapHealthMax = {}
do
	local lifeTapHealth = {
		[1454] = -20,-- rank 1
		[1455] = -60,-- rank 2
		[1456] = -130,-- rank 3
		[11687] = -210,-- rank 4
		[11688] = -300,-- rank 5
		[11689] = -420,-- rank 6
		[27222] = -580,-- rank 7
	}
	local lifeTapLevel = {
		[1454] = 6,-- rank 1
		[1455] = 16,-- rank 2
		[1456] = 26,-- rank 3
		[11687] = 36,-- rank 4
		[11688] = 46,-- rank 5
		[11689] = 56,-- rank 6
		[27222] = 68,-- rank 7
	}
	for spellID,baseCost in pairs(lifeTapHealth) do
		-- maxLevel for rank 1 is 16, rank 7 is 70.
		local maxLevel = math.min(lifeTapLevel[spellID]+10,70)
		-- maxBonus for rank 1 is 10, rank 7 is 2.
		local maxBonus = maxLevel - lifeTapLevel[spellID]
		-- rank 1 -30, rank 7 -582.
		lifeTapHealthMax[spellID] = baseCost - maxBonus
	end
end

local flagRaid = bit.bor(COMBATLOG_OBJECT_AFFILIATION_MINE, COMBATLOG_OBJECT_AFFILIATION_PARTY, COMBATLOG_OBJECT_AFFILIATION_RAID)
local flagPlayerPets = bit.bor(COMBATLOG_OBJECT_TYPE_PLAYER, COMBATLOG_OBJECT_TYPE_PET)
local bitband = bit.band
function eventHandlers.COMBAT_LOG_EVENT_UNFILTERED(timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, ...)
	local amount
	local amountMax
	local guid
	
	--local playerGUID = UnitGUID('player')
	--if destGUID == playerGUID then
	if bitband(destFlags, flagRaid)>0 and bitband(destFlags, flagPlayerPets)>0 then
		-- We only care about damage or healing, check for either of those and react accordingly
		if(event == "SWING_DAMAGE") then --autoattack dmg
			amount = -(...) -- putting in braces will autoselect the first arg, no need to use select(1, ...)
		elseif(event == "SPELL_PERIODIC_DAMAGE" or event == "SPELL_DAMAGE"
			or event == "DAMAGE_SPLIT" or event == "DAMAGE_SHIELD") then -- all kinds of spelldamage
			amount = -select(4, ...)
		elseif(event == "SPELL_HEAL" or event == "SPELL_PERIODIC_HEAL") then --healing
			amount = select(4, ...)
		elseif(event == "ENVIRONMENTAL_DAMAGE") then --environmental damage
			amount = -select(2, ...)
--[[
		else
			local applied = event == "SPELL_AURA_APPLIED"
			if applied or event == "SPELL_AURA_REMOVED" then
				local spellID = (...)
				if spellID == 469 then
					-- TODO Commanding Shout
					local health = healthFromGUID[destGUID]
					if health then
						local healthMax = healthMaxFromGUID[destGUID]
						local p = health/healthMax
						if applied then
							amount = math.floor((healthMax+1080)*p + 0.5)
							amountMax = 1080
						else
							amount = math.floor((healthMax-1080)*p + 0.5)
							amountMax = -1080
						end
					end
				end
			end
--]]
		end
		guid = destGUID
	-- elseif sourceGUID == playerGUID then
	elseif bitband(sourceFlags, flagRaid)>0 and bitband(sourceFlags, flagPlayerPets)>0 then
		if event == "SPELL_CAST_SUCCESS" then
			local spellID = (...)
			-- TODO Check only for locks?
			local lifeTapMax = lifeTapHealthMax[spellID]
			if lifeTapMax then
				amount = lifeTapMax
--[[
			elseif spellID==2687 then
				-- TODO Bloodrage
				amount = -711
--]]
			elseif spellID==12975 then
				-- Last Stand; this should be pretty exact because healthMax doesn't change
				local healthMax = healthMaxFromGUID[sourceGUID]
				if healthMax then
					amount = math.floor(0.3 * healthMax + 0.5)
					amountMax = amount
				end
			end
		end
		guid = sourceGUID
	end

	if amount then
		if HealthUpdatedDebugRegistered then
			QuickHealth.events:Fire("HealthUpdatedDebug", guid, 3, GetTime(), GetFrameNumber(), amount, amountMax)
		end
		processCLEU(guid, amount, amountMax)
	end
end
